from cstr_model import CSTRParameters, setup_cstr_model_equations
import numpy as np
from cstr_utils import plot_cstr
import time
from integrators import RK4Integrator, CVODESIntegrator
from mpc_solver import SolverCasadiMultipleShooting, SolverCasadiCollocation, SolverAcados


def simulate_closed_loop(integrator_plant, mpc_solver, x0: np.ndarray, Nsim: int):

    nx = mpc_solver.nx
    nu = mpc_solver.nu

    X = np.ndarray((Nsim + 1, nx))
    U = np.ndarray((Nsim, nu))

    closed_loop_cost = 0

    x_current = x0
    X[[0]] = x_current

    start = time.time()

    for i in range(Nsim):

        u_current = mpc_solver.solve(x_current)
        x_next = integrator_plant.simulate(x_current, u_current).T
        closed_loop_cost += mpc_solver.eval_stage_cost(x_current, u_current)
        x_current = x_next
        X[[i+1], :] = x_next
        U[[i], :] = u_current.T

    closed_loop_cost += mpc_solver.eval_terminal_cost(x_current)
    simulation_time = time.time() - start

    return X, U, simulation_time, closed_loop_cost.item()


def main():

    Nsim = 50
    N = 15 # horizon
    dt = 0.25  # [min]

    # initial state
    params = CSTRParameters()
    xs = params.xs
    us = params.us

    x0 = np.array([[0.05, 0.75, 0.5]]) * xs

    x, u, xdot, f = setup_cstr_model_equations(params)
    Q = np.diag([1.5, 1e-4, 2.])
    R = np.diag([1e-3, 10.])
    P = np.diag([1, 1e-4, 2.5])

    # TODO add your implementation here / part 3.1
    # l =
    # E =

    u_min = np.transpose(np.array([0.95, 0.85]) * us)
    u_max = np.transpose(np.array([1.05, 1.15]) * us)

    x_min = np.array([0, 0, 0.05])
    x_max = np.array([50, 600, 2.])

    X_all = []
    U_all = []
    labels_all = []
    timings_all = []
    costs_all = []

    # collocation
    integrator_plant = CVODESIntegrator(x, u, f, dt)
    num_steps = 1
    mpc_solver = SolverCasadiCollocation(x, u, f, l, E, dt, N, u_min, u_max, x_min, x_max, xs.T, us.T, dt, num_steps=num_steps)
    X, U, simulation_time, cost = simulate_closed_loop(integrator_plant, mpc_solver, x0, Nsim)

    X_all.append(X)
    U_all.append(U)
    labels_all.append(f'collocation, implicit euler, {num_steps=}')
    costs_all.append(cost)
    timings_all.append(simulation_time)


    # multiple shooting
    integrator_plant = CVODESIntegrator(x, u, f, dt)
    num_steps = 10
    model_rk4 = RK4Integrator(x, u, f, dt, num_steps=num_steps)
    f_discrete = model_rk4.f_discrete_fun(x, u)
    mpc_solver = SolverCasadiMultipleShooting(x, u, f_discrete, l, E, dt, N, u_min, u_max, x_min, x_max, xs.T, us.T)
    X, U, simulation_time, cost = simulate_closed_loop(integrator_plant, mpc_solver, x0, Nsim)

    X_all.append(X)
    U_all.append(U)
    labels_all.append(f'multiple shooting, RK4, {num_steps=}')
    costs_all.append(cost)
    timings_all.append(simulation_time)

    # TODO uncomment this part / part 3.6
    # # multiple shooting using acados
    # integrator_plant = CVODESIntegrator(x, u, f, dt)
    # mpc_solver = SolverAcados(x, u, xdot, f, Q, R, P, xs.T, us.T, u_min, u_max, x_min, x_max, dt, N)
    # X, U, simulation_time, cost = simulate_closed_loop(integrator_plant, mpc_solver, x0, Nsim)

    # X_all.append(X)
    # U_all.append(U)
    # labels_all.append(f'multiple shooting, acados')
    # costs_all.append(cost)
    # timings_all.append(simulation_time)


    # plot results
    plot_cstr(dt, X_all, U_all, labels_all)

    print("\n\nEVALUATION:")
    for l, t, c in zip(labels_all, timings_all, costs_all):
        print(f'{l}: \t total time {t:.3f} s, closed-loop cost {c:.3f}\n')


if __name__ == "__main__":
    main()
