import numpy as np
import casadi as ca
from acados_template import AcadosModel
from acados_template import AcadosOcp, AcadosOcpSolver
import scipy


class SolverCasadiCollocation():

    def __init__(self, x: ca.MX, u: ca.MX, f: ca.MX, l: ca.MX, E: ca.MX, dt: float, N: int,
                 u_min: np.ndarray, u_max: np.ndarray, x_min: np.ndarray, x_max: np.ndarray, xs: np.ndarray, us:np.ndarray, delta_t: float, num_steps: int):

        self.nx = x.shape[0]
        self.nu = u.shape[0]

        self.l_fun = ca.Function('stage_cost', [x, u], [l])
        self.E_fun = ca.Function('terminal_cost', [x], [E])

        f_fun = ca.Function('f', [x, u], [f])

        x0_bar = ca.MX.sym('x0_bar', x.shape)

        objective = 0

        x_current = ca.MX.sym('x_0', x.shape)
        variables = [x_current]
        g = [x_current - x0_bar]
        lb_vars = [x_min]
        ub_vars = [x_max]

        vars_initial_guess = [xs]

        h = delta_t/num_steps

        for n in range(N):
            u_current = ca.MX.sym(f'u_{n}', u.shape)

            variables.append(u_current)
            vars_initial_guess.append(us)
            lb_vars.append(u_min)
            ub_vars.append(u_max)

            objective += self.l_fun(x_current, u_current)

            for i in range(num_steps):
                k_current = ca.MX.sym(f'k_{n}_{i}', x.shape)
                x_next = ca.MX.sym(f'x_{n+1}_{i}', x.shape)

                # collocation equations
                # TODO add your implementation here / part 2.2
                # append the collocation equations to g, in the OCP we will have g = 0

                variables.append(k_current)
                vars_initial_guess.append(np.zeros(k_current.shape))

                lb_vars.append(-np.inf*np.ones(k_current.shape))
                ub_vars.append(np.inf*np.ones(k_current.shape))

                variables.append(x_next)
                vars_initial_guess.append(xs)
                lb_vars.append(x_min)
                ub_vars.append(x_max)

                x_current = x_next

        objective += self.E_fun(x_current)

        g = ca.vertcat(*g)
        self.lbg = np.zeros(g.shape)
        self.ubg = np.zeros(g.shape)
        self.lb_vars = ca.vertcat(*lb_vars)
        self.ub_vars = ca.vertcat(*ub_vars)
        variables = ca.vertcat(*variables)

        nlp = {'f': objective, 'x': variables, 'g': g, 'p': x0_bar}
        self.solver = ca.nlpsol('solver', 'ipopt', nlp)


        self.sol_guess = ca.vertcat(*vars_initial_guess)


    def solve(self, x0):

        res = self.solver(x0=self.sol_guess, p=x0, lbg=self.lbg, ubg=self.ubg, lbx=self.lb_vars, ubx=self.ub_vars)

        w_opt = res['x'].full()
        self.sol_guess = w_opt

        return w_opt[self.nx:self.nx+self.nu]


    def eval_stage_cost(self, x, u):
        return self.l_fun(x, u).full()


    def eval_terminal_cost(self, x):
        return self.E_fun(x).full()



class SolverCasadiMultipleShooting():

    def __init__(self, x: ca.MX, u: ca.MX, f: ca.MX, l: ca.MX, E: ca.MX, dt: float, N: int,
                 u_min: np.ndarray, u_max: np.ndarray, x_min: np.ndarray, x_max: np.ndarray, xs: np.ndarray, us:np.ndarray):

        self.nx = x.shape[0]
        self.nu = u.shape[0]

        self.l_fun = ca.Function('stage_cost', [x, u], [l])
        self.E_fun = ca.Function('terminal_cost', [x], [E])

        f_fun = ca.Function('f', [x, u], [f])

        x0_bar = ca.MX.sym('x0_bar', x.shape)

        objective = 0

        x_current = ca.MX.sym('x_0', x.shape)
        variables = [x_current]
        g = [x_current - x0_bar]
        lb_vars = [x_min]
        ub_vars = [x_max]

        vars_initial_guess = [xs]

        for n in range(N):
            u_current = ca.MX.sym(f'u_{n}', u.shape)
            x_next = ca.MX.sym(f'x_{n+1}', x.shape)

            objective += self.l_fun(x_current, u_current)

            # TODO add your implementation here / part 2.4
            # add the dynamics constraint to g

            variables.append(u_current)
            lb_vars.append(u_min)
            ub_vars.append(u_max)

            variables.append(x_next)
            lb_vars.append(x_min)
            ub_vars.append(x_max)

            vars_initial_guess.append(us)
            vars_initial_guess.append(xs)

            x_current = x_next

        objective += self.E_fun(x_current)

        g = ca.vertcat(*g)
        self.lbg = np.zeros((self.nx*(N+1), 1))
        self.ubg = np.zeros((self.nx*(N+1), 1))
        self.lb_vars = ca.vertcat(*lb_vars)
        self.ub_vars = ca.vertcat(*ub_vars)
        variables = ca.vertcat(*variables)

        nlp = {'f': objective, 'x': variables, 'g': g, 'p': x0_bar}
        self.solver = ca.nlpsol('solver', 'ipopt', nlp)

        self.sol_guess = ca.vertcat(*vars_initial_guess)


    def solve(self, x0):

        res = self.solver(x0=self.sol_guess, p=x0, lbg=self.lbg, ubg=self.ubg, lbx=self.lb_vars, ubx=self.ub_vars)

        w_opt = res['x'].full()
        self.sol_guess = w_opt

        return w_opt[self.nx:self.nx+self.nu]


    def eval_stage_cost(self, x, u):
        return self.l_fun(x, u).full()


    def eval_terminal_cost(self, x):
        return self.E_fun(x).full()



class SolverAcados():

    def __init__(self, x: ca.MX, u: ca.MX, xdot: ca.MX, f: ca.MX, Q: np.ndarray, R: np.ndarray, P: np.ndarray, xs: np.ndarray, us:np.ndarray, 
                 u_min: np.ndarray, u_max: np.ndarray, x_min: np.ndarray, x_max: np.ndarray, dt: float, N: int):

        self.nx = x.shape[0]
        self.nu = u.shape[0]

        # this is just needed for the closed-loop simulation
        l = (x-xs).T @ Q @ (x-xs) + (u - us).T @ R @ (u - us)
        E = (x-xs).T @ P @ (x-xs)
        self.l_fun = ca.Function('stage_cost', [x, u], [l])
        self.E_fun = ca.Function('terminal_cost', [x], [E])

        # setting up the acados model
        model = AcadosModel()

        f_impl = xdot - f
        model.f_impl_expr = f_impl
        model.f_expl_expr = f
        model.x = x
        model.xdot = xdot
        model.u = u
        model.name = 'cstr'

        self.model = model

        # setting up the acados OCP
        ocp = AcadosOcp()

        ocp.model = self.model
        ocp.dims.N = N
        ocp.solver_options.tf = N*dt

        # set cost
        ocp.cost.W = 1/dt*scipy.linalg.block_diag(Q, R)
        ocp.cost.W_e = P

        ocp.cost.cost_type = "NONLINEAR_LS"
        ocp.cost.cost_type_e = "NONLINEAR_LS"

        ocp.model.cost_y_expr = ca.vertcat(self.model.x, self.model.u)
        ocp.model.cost_y_expr_e = self.model.x

        ocp.cost.yref = ca.vertcat(xs, us).full().ravel()
        ocp.cost.yref_e = xs.ravel()

        # set constraints
        ocp.constraints.lbu = u_min
        ocp.constraints.ubu = u_max
        ocp.constraints.idxbu = np.arange(self.nu)

        ocp.constraints.lbx = x_min
        ocp.constraints.ubx = x_max
        ocp.constraints.idxbx = np.arange(self.nx)

        ocp.constraints.x0 = xs

        # set options
        ocp.solver_options.qp_solver = "PARTIAL_CONDENSING_HPIPM"  # FULL_CONDENSING_QPOASES
        # PARTIAL_CONDENSING_HPIPM, FULL_CONDENSING_QPOASES, FULL_CONDENSING_HPIPM,
        ocp.solver_options.qp_solver_cond_N = N  # for partial condensing

        # ocp.solver_options.print_level = 1
        ocp.solver_options.hessian_approx = "GAUSS_NEWTON"
        ocp.solver_options.nlp_solver_type = "SQP"  # SQP_RTI, SQP
        ocp.solver_options.levenberg_marquardt = 1e-5

        # integrator options
        ocp.solver_options.integrator_type = "IRK"
        ocp.solver_options.sim_method_num_stages = 4
        ocp.solver_options.sim_method_num_steps = 1
        ocp.solver_options.sim_method_newton_iter = 10

        self.solver = AcadosOcpSolver(ocp, json_file="acados_ocp.json")

        # set initial guess
        for n in range(N):
            self.solver.set(n, 'x', xs)
            self.solver.set(n, 'u', us)
        self.solver.set(N, 'x', xs)


    def solve(self, x0):
        u0 = self.solver.solve_for_x0(x0.ravel(), fail_on_nonzero_status=False)
        return u0


    def eval_stage_cost(self, x, u):
        return self.l_fun(x, u).full()


    def eval_terminal_cost(self, x):
        return self.E_fun(x).full()